/* ivac.c -- main ivac file
 *
 * author: hackbard@hackdaworld.dyndns.org
 *         frank.zirkelbach@physik.uni-augsburg.de
 *
 * Copyright (C) 2004 Frank Zirkelbach
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// #define USE_NCURSES

#include "ivac.h"

int main(int argc,char **argv) {

  /* TESTING BY NOW */

  t_ivac ivac;
  int i;

  if(argc!=2) {
    printf("\nusage: %s <username>\n\n",argv[0]);
    return ERROR;
  }

  /* set username */
  strncpy(ivac.username,argv[1],CHAR_USERNAME);
  for(i=0;i<IVAC_CONSOLE_LEN;i++)
    memset(ivac.console[i],0,IVAC_CONSOLE_STRING_LEN);

  /* set capabilities (futur: set by check routines) */
  ivac.g_cap=NETWORK;
  ivac.av_cap=AUDIO|VIDEO|DUPLEX;

  /* set event timeout */
  ivac.event.timeout.tv_sec=IVAC_S_SEC;
  ivac.event.timeout.tv_usec=IVAC_S_USEC;
  /* event init */
  event_init(&(ivac.event));

  /* input init */
  ivac.input.mode=CONTENT_BUFFER;
  input_init(&(ivac.input));

  /* set listen port (futur: read from config or entered later) */
  network_set_listen_port(&(ivac.net),IVAC_LISTEN_PORT);
  /* network init */
  if(network_init(&(ivac.net))==N_ERROR) {
    printf("[ivac] use 'fuser -n tcp %d' to determine the process to kill!\n",
           ivac.net.l_port);
    ivac_shutdown(&ivac);
    return ERROR;
  }

  /* add listening port + stdin to (read) event system */
  event_math(ivac.net.l_fd,&(ivac.event),READ,ADD);
  event_math(0,&(ivac.event),READ,ADD);

  /* display init */
  display_init(&(ivac.display));

  /* display */
  ivac_display(&(ivac.display));

  /* start event system - callbacks used: ivac_event_cb + ivac_regular_cb */
  event_start(&(ivac.event),(void *)&ivac,ivac_event_cb,ivac_regular_cb);

  // ivac_shutdown(&ivac);

  return SUCCESS;
}

int ivac_shutdown(t_ivac *ivac) {

  network_shutdown(&(ivac->net));
  input_shutdown(&(ivac->input));
  event_stop(&(ivac->event));
#ifdef USE_NCURSES
  display_shutdown(&(ivac->display));
#endif

  return SUCCESS;
}

int ivac_send_info(int channel,t_ivac *ivac) {

  char data[SEND_N_MAX];
  int size;

  size=strlen(ivac->username);

  data[0]=IVAC_SEND_NAME;
  data[1]=size;
  strncpy(data+2,ivac->username,size);
  size+=2;

  data[size]=IVAC_SEND_G_CAP;
  data[size+1]=1;
  data[size+2]=ivac->g_cap;
  size+=3;

  data[size]=IVAC_SEND_AV_CAP;
  data[size+1]=2;
  data[size+2]=(ivac->av_cap)>>8;
  data[size+3]=(ivac->av_cap)&0xff;
  size+=4;

  if(network_send(ivac->net.connection[channel].fd,data,size)==N_ERROR) {
    puts("[ivac] ivac_send_info failed");
    return ERROR;
  }

  return SUCCESS;
}

int ivac_receive_info(int channel,t_ivac *ivac) {

  char data[SEND_N_MAX];
  int count,length;

  count=0;

  if((length=network_receive(ivac->net.connection[channel].fd,
                             data,SEND_N_MAX))==N_ERROR) {
    puts("[ivac] ivac_receive_info failed");
    return ERROR;
  }

  while(length-count) {
    switch(data[count]) {
      case IVAC_SEND_NAME:
        strncpy(ivac->challenger[channel].name,data+count+2,data[count+1]);
        ivac->challenger[channel].name[data[count+1]]='\0';
        count+=(data[count+1]+2);
        break;
      case IVAC_SEND_G_CAP:
        ivac->challenger[channel].g_cap=data[count+2];
        count+=3;
        break;
      case IVAC_SEND_AV_CAP:
        ivac->challenger[channel].av_cap=data[count+2]<<8;
        ivac->challenger[channel].av_cap|=data[count+3];
        count+=4;
        break;
      default:
        puts("[ivac] ivac_receive_info, unknown character");
        return ERROR;
        break;
    }
  }

  return SUCCESS;
}

int ivac_event_cb(t_event *event,void *ptr) {
 
  t_ivac *ivac;
  int channel;

  ivac=(t_ivac *)ptr;

  if(FD_ISSET(ivac->net.l_fd,&(event->rfds))) {
    /* manage incoming + send info */
    channel=network_manage_incoming(&(ivac->net));
    event_math(ivac->net.connection[channel].fd,event,READ,ADD);
    ivac_send_info(channel,ivac);
  }

  /* receive info */
  for(channel=0;channel<MAX_CONNECTIONS;channel++)
    if(ivac->net.connection[channel].status&C_ESTABL)
      if(FD_ISSET(ivac->net.connection[channel].fd,&(event->rfds)))
        ivac_receive_info(channel,ivac);

  /* user interaction */
  if(FD_ISSET(0,&(event->rfds)))
    input_get_event(&(ivac->input),ivac_parse_command,ivac);

  /* display ivac gui */
#ifdef USE_NCURSES
  ivac_display_content(ivac);
#endif
    
  return SUCCESS;
}

int ivac_regular_cb(t_event *event,void *ptr) {

  /* usual jobs like audio & video transmit ... */

  return SUCCESS;
}

int ivac_parse_command(t_input *input,void *ptr) {

  t_ivac *ivac;
  int i,j,k;
  int len;
  int channel;
  char *data,valid;
  char c_str[IVAC_CONSOLE_STRING_LEN];
  char arg[IVAC_ARG_COUNT][IVAC_ARG_LEN];

  ivac=(t_ivac *)ptr;
  data=input->content;
  valid=0;

  printf("%c",data[input->c_count-1]);
  fflush(NULL);

  /* parse command routines */

  if(data[input->c_count-1]=='\n') {

    /* delete console string + args */
    memset(c_str,0,IVAC_CONSOLE_STRING_LEN);
    for(j=0;j<IVAC_ARG_COUNT;j++) memset(arg[j],0,IVAC_ARG_LEN);

    /* get args */

    len=0;
    while(data[len]!='\n') len++;
    i=0; j=0;
    while((i<len) && (j<IVAC_ARG_COUNT)) {
      k=0;
      while((data[i+k]!=' ')&&(data[i+k]!='\n')) {
        arg[j][k]=data[i+k];
        k++;
      }
      arg[j][k]='\0';
      j++;
      /* skip all ' ' */
      while(data[i+k]==' ') k++;
      i+=k;
    }

    /* parse command  aka arg[0] */
    if(!(strncmp(arg[0],"quit",4))) {
      valid=1;
      sprintf(c_str,"ivac shutdown ...");
      ivac_shutdown(ivac);
    }
    if(!(strncmp(arg[0],"set",3))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if(!(strncmp(arg[1],"name",4))) {
        strncpy(ivac->username,arg[2],CHAR_USERNAME);
        sprintf(c_str,"changed username to %s",ivac->username);
      }
      else if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        if(ivac->net.connection[channel].status&C_IN_USE) {
          sprintf(c_str,"channel %02d: connection in use",channel);
        }
        else {
          strncpy(ivac->net.connection[channel].ip,arg[2],IP_DIGITS);
          ivac->net.connection[channel].port=atoi(arg[3]);
          sprintf(c_str,"channel %02d: set connection info",channel);
          ivac->net.connection[channel].status|=C_INFO_A;
        }
      } else snprintf(c_str,IVAC_CONSOLE_STRING_LEN,"unknown argument: '%s'",
                      arg[1]);
    }
    if(!(strncmp(arg[0],"connect",7))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        if(!(ivac->net.connection[channel].status&C_INFO_A))
          sprintf(c_str,"channel %02d: channel not configured",channel);
        else {
          if(ivac->net.connection[channel].status&C_IN_USE)
            sprintf(c_str,"channel %02d: connection in use",channel);
          else {
            sprintf(c_str,"channel %02d: trying to connect to %s:%d",channel,
                    ivac->net.connection[channel].ip,
                    ivac->net.connection[channel].port);
          }
        }
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    } 
    if(!(strncmp(arg[0],"close",5))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        if(!(ivac->net.connection[channel].status&C_ESTABL))
          sprintf(c_str,"channel %02d: no active connection",channel);
        else {
          ivac->net.connection[channel].status|=C_HANGUP;
          sprintf(c_str,"channel %02d: connection closed",channel);
        }
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    }
    if(!(strncmp(arg[0],"select",6))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        ivac->net.sendmask|=(1<<channel);
        sprintf(c_str,"selected channel %d",channel);
      }
      else if(arg[1][0]='*') {
        ivac->net.sendmask=0xff;
        strcpy(c_str,"selected all channels");
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    }
    if(!(strncmp(arg[0],"deselect",8))) {
      valid=1;
      if((arg[1][0]>='0')&&(arg[1][0]<='9')) channel=atoi(arg[1]);
      else channel=-1;
      if((channel>=0)&&(channel<MAX_CONNECTIONS)) {
        ivac->net.sendmask&=(~(1<<channel));
        sprintf(c_str,"deselected channel %d",channel);
      }
      else if(arg[1][0]='*') {
        ivac->net.sendmask=0;
        strcpy(c_str,"deselected all channels");
      }
      else sprintf(c_str,"invalid argument: '%s'",arg[1]);
    }

    if(!valid)
      snprintf(c_str,IVAC_CONSOLE_STRING_LEN,"unknown command: '%s'",arg[0]);

    ivac_add_to_monitor(ivac,c_str);
      

    /* delete content buffer + reset counter */
    memset(input->content,0,input->c_count-1);
    input->c_count=0;

  }

  return SUCCESS;
}

int ivac_display_head(t_display *display) {

#ifdef USE_NCURSES
  int x,y;

  move(0,0);
  for(x=0;x<display->max_x;x++) addch('#');
  mvaddstr(1,0,"##");
  mvaddstr(1,(display->max_x-4)/2-4,"- ivac -");
  mvaddstr(1,(display->max_x-2),"##");
  move(2,0);
  for(x=0;x<display->max_x;x++) addch('#');
  refresh();
#else
  puts("#########################################################");
  puts("##### ivac - -  Copyright (C) 2004 Frank Zirkelbach #####");
  puts("#########################################################");
#endif

  return SUCCESS;
}

int ivac_display_box(t_display *display) {
 
#ifdef USE_NCURSES 
  int x,y;

  for(y=IVAC_PROMPT_LEN;y<display->max_y-IVAC_PROMPT_LEN;y++) {
    mvaddch(y,0,'#');
    mvaddch(y,display->max_x-1,'#');
  }
#endif

  return SUCCESS;
}

int ivac_display_box_content(t_ivac *ivac) {


  return SUCCESS;
}

int ivac_display_console(t_display *display) {
                                             
  int x,y;
         
  move(display->max_y-IVAC_CONSOLE_LEN-IVAC_PROMPT_LEN-1,0);
  for(x=0;x<display->max_x;x++) addch('#');                         
                                          
  return SUCCESS;
}

int ivac_display_console_content(t_ivac *ivac) {

  int x,y;
  int len;

  for(y=0;y<IVAC_CONSOLE_LEN;y++) {
    len=strlen(ivac->console[y]);
    move(ivac->display.max_y-IVAC_CONSOLE_LEN-IVAC_PROMPT_LEN+y,2);
    for(x=0;x<len;x++)
       addch(((ivac->console[y][x]>' ')||(ivac->console[y][x]<='~'))
               ?ivac->console[y][x]:' ');
    for(x=len;x<IVAC_CONSOLE_STRING_LEN-4;x++) addch(' ');
  }
  refresh();

  return SUCCESS;
}

int ivac_display_prompt(t_display *display) {

  int x,y;

  move(display->max_y-3,0);
  for(x=0;x<display->max_x;x++) addch('#');
  mvaddstr(display->max_y-2,0,"## command: ");
  mvaddstr(display->max_y-2,display->max_x-2,"##");
  move(display->max_y-1,0);
  for(x=0;x<display->max_x;x++) addch('#');
  refresh();

  return SUCCESS;
}

int ivac_display_prompt_content(t_ivac *ivac) {

  int x,y;

  /* delete old command */
  if(ivac->input.c_count==0) {
    move(ivac->display.max_y-2,12);
    for(x=12;x<ivac->display.max_x-1;x++) addch(' ');
  }

  for(x=0;x<ivac->input.c_count;x++)
    mvaddch(ivac->display.max_y-2,x+12,ivac->input.content[x]);
  refresh();

  return SUCCESS;
}

int ivac_display(t_display *display) {

  int x,y;

  /* display head */
  ivac_display_head(display);

  /* display box */
  ivac_display_box(display);

  /* display console */
  ivac_display_console(display);

  /* display command prompt */
  ivac_display_prompt(display);

  return SUCCESS;
}

int ivac_display_content(t_ivac *ivac) {

  /* display box content */
  ivac_display_box_content(ivac);

  /* display console content */
  ivac_display_console_content(ivac);

  /* display prompt content */
  ivac_display_prompt_content(ivac);

  return SUCCESS;
}

int ivac_add_to_monitor(t_ivac *ivac,char *msg) {

  int i;

  for(i=0;i<IVAC_CONSOLE_LEN-1;i++)
    memcpy(ivac->console[i],ivac->console[i+1],IVAC_CONSOLE_STRING_LEN);
  memcpy(ivac->console[IVAC_CONSOLE_LEN-1],msg,IVAC_CONSOLE_STRING_LEN);

  for(i=0;i<IVAC_CONSOLE_LEN;i++)
    printf("[ivac] console line %d: %s\n",i,ivac->console[i]);
#ifdef USE_NCURSES
  ivac_display_console_content(ivac);
#endif

  return SUCCESS;
}
